import logging

from django.db.models.signals import pre_delete
from django.dispatch import receiver

from sidekick.models import CronSchedule
from sidekick.services.cron_schedule import CronScheduleService

logger = logging.getLogger(__name__)


@receiver(pre_delete, sender=CronSchedule)
def cron_schedule_pre_delete(sender, instance, **kwargs):
    """Methods to be run on the pre delete of a CronSchedule """
    # noinspection PyBroadException
    try:
        CronScheduleService.handle_removing_schedule(instance)
    except Exception:
        logger.error(msg="Failed during cron schedule pre delete")



import logging
from datetime import datetime, timezone

from django.db import models

logger = logging.getLogger(__name__)


class Task(models.Model):
    """Task model used to define tasks that will be run as cron jobs """

    FAILED = 'FAILED'
    IN_PROGRESS = 'IN_PROGRESS'
    SLEEPING = 'SLEEPING'
    SUCCESS = 'SUCCESS'

    CURRENT_STATUS = (
        (FAILED, 'Failed'),
        (IN_PROGRESS, 'In Progress'),
        (SLEEPING, 'Sleeping'),
        (SUCCESS, 'Success')
    )

    name = models.CharField(max_length=255, help_text='The human readable name of the task')
    registered_task = models.ForeignKey('sidekick.RegisteredTask', null=True, blank=True, on_delete=models.CASCADE)
    cron_schedule = models.ForeignKey('sidekick.CronSchedule', null=True, blank=True, on_delete=models.DO_NOTHING)
    enabled = models.BooleanField(default=False, help_text='Whether the task is enabled')
    status = models.CharField(max_length=11, choices=CURRENT_STATUS, default=SLEEPING)
    started_at = models.DateTimeField(null=True, blank=True, timezone=timezone.utc)
    finished_at = models.DateTimeField(null=True, blank=True, timezone=timezone.utc)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        """Task Pre Save"""
        if not self.registered_task or not self.cron_schedule:
            self.enabled = False

        super(Task, self).save(*args, **kwargs)

        """Task Post Save"""
        try:
            from sidekick.services.cron_files import CronService
            CronService().generate_cron_tasks()
        except Exception as e:
            logger.error(msg='Failed to generate cron task for {} due to {}'.format(self, e))

    def task_of(self):
        """The app which this task belongs to """
        return self.registered_task.task_name.split(' ')[0]

    def running_for(self):
        """How long has the current task been running for"""
        if self.status == self.IN_PROGRESS:
            now = datetime.now(timezone.utc)
            difference = now - self.started_at
            prefix = "Running for"
        elif self.started_at and self.finished_at:
            difference = self.finished_at - self.started_at
            prefix = "Completed in"
        else:
            return "Not currently running."
        days, seconds = difference.days, difference.seconds
        hours = days * 24 + seconds // 3600
        minutes = (seconds % 3600) // 60
        seconds = seconds % 60
        return "{} {} hours, {} minutes, {} seconds".format(prefix, hours, minutes, seconds)


class CronSchedule(models.Model):
    """Cron Schedule model for defining different time intervals """

    name = models.CharField(max_length=100, help_text='Human readable version of schedule, i.e Every 10 minutes')
    minute = models.CharField(max_length=100, help_text='At what minute. * for every', default='*')
    hour = models.CharField(max_length=100, help_text='At what hour. * for every', default='*')
    day_of_week = models.CharField(max_length=100, help_text='At what day of the week. * for every', default='*')
    day_of_month = models.CharField(max_length=100, help_text='At what day of the month. * for every', default='*')
    month_of_year = models.CharField(max_length=100, help_text='At what month of the year. * for every', default='*')

    def __str__(self):
        return self.name

    def schedule(self):
        return "{} {} {} {} {}".format(self.minute, self.hour, self.day_of_month, self.month_of_year, self.day_of_week)


class RegisteredTask(models.Model):
    """Model for the registered task to be used by the Task Model """

    task_name = models.CharField(max_length=125, help_text='The task to be run ie. stock --get_stock_updates')

    def __str__(self):
        return self.task_name.replace('--', ' - ').replace('_', ' ')



import logging
from sidekick.models import Task

logger = logging.getLogger(__name__)


class CronScheduleService:
    """
    Services for the cron schedule
    """
    @staticmethod
    def handle_removing_schedule(schedule_instance):
        """
        If a cron schedule is deleted then make sure you mark any tasks that were using it as no longer enabled
        :return:
        """
        logger.info(msg="Pre Delete of CronSchedule instance - disabling tasks with this cron schedule")
        tasks = Task.objects.filter(cron_schedule=schedule_instance)
        # noinspection PyBroadException
        try:
            for task in tasks:
                task.enabled = False
                task.save()
                logger.info(msg="{} has been disabled due to no longer having a cron schedule associated to it".format(
                    task
                ))
        except Exception:
            logger.error(msg="Failed during process of disabling sidekick tasks on pre_delete of a cron schedule")



"""
Helper functions to make the code quicker to write and easier for others to read
"""
from datetime import datetime, timezone
from sidekick.models import Task


def get_task_name(options):
    """
    Given a dictionary of command options, return the name of the task
    :param options: Options passed to the handle method of the management command
    :return: The task name (str)
    """
    options_dict = dict(options)
    return [task for task in [key for key in options_dict] if str(options_dict[task]) == 'True'][0]


def get_app_name(name):
    """
    Given the __name__ return the app name
    :param name: __name__
    :return: App name
    """
    return name.split('.')[-1]


def update_task_status(task_name, status):
    """Update the status, started_at and finished_at fields"""

    if status == Task.IN_PROGRESS:
        Task.objects.filter(name=task_name).update(status=status, started_at=datetime.now(timezone.utc))
    elif status in {Task.SUCCESS, Task.FAILED}:
        Task.objects.filter(name=task_name).update(status=status, finished_at=datetime.now(timezone.utc))
    else:
        Task.objects.filter(name=task_name).update(status=status, started_at=None, finished_at=None)



from django.db import migrations


def do_func(apps, schema_editor):
    # We get the model from the versioned app registry;
    # if we directly import it, it'll be the wrong version
    CronSchedule = apps.get_model("sidekick", "CronSchedule")
    db_alias = schema_editor.connection.alias
    CronSchedule.objects.using(db_alias).bulk_create([
        CronSchedule(name="Every 5 minutes all day",
                     minute="*/5",
                     hour="*",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     ),
        CronSchedule(name="Every 10 minutes all day",
                     minute="*/10",
                     hour="*",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     ),
        CronSchedule(name="Every 15 minutes all day",
                     minute="*/15",
                     hour="*",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     ),
        CronSchedule(name="Every 30 minutes all day",
                     minute="*/30",
                     hour="*",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     ),
        CronSchedule(name="Every hour, all day",
                     minute="0",
                     hour="*",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     ),
        CronSchedule(name="At 5pm every day",
                     minute="0",
                     hour="17",
                     day_of_week="*",
                     day_of_month="*",
                     month_of_year="*"
                     )
    ])


def undo_func(apps, schema_editor):
    CronSchedule = apps.get_model("sidekick", "CronSchedule")
    db_alias = schema_editor.connection.alias
    CronSchedule.objects.using(db_alias).filter(name="Every 5 minutes all day").delete()
    CronSchedule.objects.using(db_alias).filter(name="Every 10 minutes all day").delete()
    CronSchedule.objects.using(db_alias).filter(name="Every 15 minutes all day").delete()
    CronSchedule.objects.using(db_alias).filter(name="Every 30 minutes all day").delete()
    CronSchedule.objects.using(db_alias).filter(name="Every hour, all day").delete()
    CronSchedule.objects.using(db_alias).filter(name="At 5pm every day").delete()


class Migration(migrations.Migration):

    dependencies = [
        ('sidekick', '0001_initial')
    ]

    operations = [
        migrations.RunPython(do_func, undo_func),
    ]



import logging

from django.conf import settings

from sidekick.models import Task
from sidekick.services.helpers import update_task_status


logger = logging.getLogger(__name__)


class CronService:
    """Service to create or re-write the cron files """

    def __init__(self):
        if all([getattr(settings, "SIDEKICK")['SIDEKICK_REGISTERED_APPS'],
                getattr(settings, "SIDEKICK")['ENVIRONMENT'],
                getattr(settings, "SIDEKICK")['DJANGO_PATH'],
                getattr(settings, "SIDEKICK")['CRON_PATH']]):
            self.environment = settings.SIDEKICK['ENVIRONMENT']
            self.joiner = "&& cd"
            self.django_path = settings.SIDEKICK['DJANGO_PATH']
        else:
            logger.error(msg="Missing one of SIDEKICK_REGISTERED_APPS, ENVIRONMENT, DJANGO_PATH or CRON_PATH in "
                             "settings file.")
            print('Make sure you have SIDEKICK_REGISTERED_APPS, ENVIRONMENT, DJANGO_PATH and CRON_PATH in your '
                  'settings file')

    def write_cron_file(self, tasks):
        """For each task in list, write to the cron file. """

        with open(settings.SIDEKICK['CRON_PATH'], 'w+') as ws_cron_file:
            for task in tasks:
                ws_cron_file.write(
                    "# {name}\n"
                    "{schedule} {environment} {joiner} {django_path} {task}\n\n".format(
                        name=task.name,
                        environment=self.environment,
                        joiner=self.joiner,
                        django_path=self.django_path,
                        schedule=task.cron_schedule.schedule(),
                        task=task.registered_task.task_name)
                )
                update_task_status(task_name=task.name, status='SLEEPING')

    def generate_cron_tasks(self):
        """Create a new cron file on the post save of a Registered Task"""
        try:
            tasks = Task.objects.filter(enabled=True)
            self.write_cron_file(tasks=tasks)
            logger.info(msg='Cron tasks successfully created.')
        except Exception as e:
            logger.error(msg='Failed to write cron tasks due to {}'.format(e))



from django.apps import AppConfig
from django.conf import settings


class SidekickConfig(AppConfig):
    name = 'sidekick'

    def ready(self):
        self.register_tasks()

    @staticmethod
    def register_tasks():
        """
        For each app listed in SIDEKICK_REGISTERED_APPS import the task to trigger the decorator
        :return:
        """
        import importlib
        try:
            for app in settings.SIDEKICK['SIDEKICK_REGISTERED_APPS']:
                app_task = app + '.tasks'
                importlib.import_module("%s" % app_task)
        except Exception as e:
            print("Failed to register tasks for Side Kick. Make sure you have SIDEKICK_REGISTERED_APPS within"
                  " the SIDEKICK dictionary.")



# Generated by Django 2.0.5 on 2019-02-12 10:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sidekick', '0002_create_common_cron'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='cron_schedule',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.DO_NOTHING, to='sidekick.CronSchedule'),
        ),
        migrations.AlterField(
            model_name='task',
            name='registered_task',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, to='sidekick.RegisteredTask'),
        ),
    ]



# Generated by Django 2.0.5 on 2018-12-20 15:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CronSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Human readable version of schedule, i.e Every 10 minutes', max_length=100)),
                ('minute', models.CharField(default='*', help_text='At what minute. * for every', max_length=100)),
                ('hour', models.CharField(default='*', help_text='At what hour. * for every', max_length=100)),
                ('day_of_week', models.CharField(default='*', help_text='At what day of the week. * for every', max_length=100)),
                ('day_of_month', models.CharField(default='*', help_text='At what day of the month. * for every', max_length=100)),
                ('month_of_year', models.CharField(default='*', help_text='At what month of the year. * for every', max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='RegisteredTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_name', models.CharField(help_text='The task to be run ie. stock --get_stock_updates', max_length=125)),
            ],
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The human readable name of the task', max_length=255)),
                ('enabled', models.BooleanField(default=False, help_text='Whether the task is enabled')),
                ('cron_schedule', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='sidekick.CronSchedule')),
                ('registered_task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='sidekick.RegisteredTask')),
            ],
        ),
    ]



from django.contrib import admin
from sidekick.models import Task, CronSchedule


class CronScheduleAdmin(admin.ModelAdmin):
    """
    Admin for the CronSchedule
    """
    model = CronSchedule
    fields = (
        'name',
        'minute',
        'hour',
        'day_of_week',
        'day_of_month',
        'month_of_year'
    )


admin.site.register(CronSchedule, CronScheduleAdmin)


class TaskAdmin(admin.ModelAdmin):
    """
    Admin for the task model
    """
    list_display = ['name', 'registered_task', 'status', 'running_for', 'cron_schedule', 'enabled']


admin.site.register(Task, TaskAdmin)



import setuptools

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="side-kick",
    version="0.1.9",
    author="Ben Currie",
    author_email="benjamin.w.currie@gmail.com",
    description="A lightweight task runner for django with easy admin controls",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/CurrieBen/sidekick",
    license="MIT License",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
